/* $VER: vlink aout.h V0.8 (21.01.01)
 *
 * This file is part of vlink, a portable linker for multiple
 * object formats.
 * Copyright (c) 1997-2001  Frank Wille
 *
 * vlink is freeware and part of the portable and retargetable ANSI C
 * compiler vbcc, copyright (c) 1995-2001 by Volker Barthelmann.
 * vlink may be freely redistributed as long as no modifications are
 * made and nothing is charged for it. Non-commercial usage is allowed
 * without any restrictions.
 * EVERY PRODUCT OR PROGRAM DERIVED DIRECTLY FROM MY SOURCE MAY NOT BE
 * SOLD COMMERCIALLY WITHOUT PERMISSION FROM THE AUTHOR.
 *
 * v0.8  (21.01.01) phx
 *       File created.
 */


/* a.out header */
struct aout_hdr {
  uint32  a_midmag;
  uint32  a_text;
  uint32  a_data;
  uint32  a_bss;
  uint32  a_syms;
  uint32  a_entry;
  uint32  a_trsize;
  uint32  a_drsize;
};

/* a_magic */
#define OMAGIC 0407    /* old impure format */
#define NMAGIC 0410    /* read-only text */
#define ZMAGIC 0413    /* demand load format */
#define QMAGIC 0314    /* not supported */

/* a_mid - machine id */
#define MID_SUN010      1       /* sun 68010/68020 binary */
#define MID_SUN020      2       /* sun 68020-only binary */
#define MID_PC386       100     /* 386 PC binary. (so quoth BFD) */
#define MID_HP200       200     /* hp200 (68010) BSD binary */
#define MID_I386        134     /* i386 BSD binary */
#define MID_M68K        135     /* m68k BSD binary with 8K page sizes */
#define MID_M68K4K      136     /* m68k BSD binary with 4K page sizes */
#define MID_NS32532     137     /* ns32532 */
#define MID_SPARC       138     /* sparc */
#define MID_PMAX        139     /* pmax */
#define MID_VAX1K       140     /* vax 1K page size binaries */
#define MID_ALPHA       141     /* Alpha BSD binary */
#define MID_MIPS        142     /* big-endian MIPS */
#define MID_ARM6        143     /* ARM6 */
#define MID_SH3         145     /* SH3 */
#define MID_POWERPC     149     /* big-endian PowerPC */
#define MID_VAX         150     /* vax */
#define MID_SPARC64     151     /* LP64 sparc */
#define MID_HP300       300     /* hp300 (68020+68881) BSD binary */
#define MID_HPUX        0x20C   /* hp200/300 HP-UX binary */
#define MID_HPUX800     0x20B   /* hp800 HP-UX binary */

/* a_flags */
#define EX_DYNAMIC      0x20
#define EX_PIC          0x10
#define EX_DPMASK       0x30

/* a_midmag macros */
#define GETMAGIC(a) ((uint32)(read32be(&(a)->a_midmag)&0xffff))
#define GETMID(a)   ((uint32)((read32be(&(a)->a_midmag)>>16)&0x3ff))
#define GETFLAGS(a) ((uint32)((read32be(&(a)->a_midmag)>>26)&0x3f))
#define SETMIDMAG(a,mag,mid,flag) write32be(&(a)->a_midmag, \
                  ((flag)&0x3f)<<26|((mid)&0x3ff)<<16|((mag)&0xffff))


/* Relocation info structures */
struct relocation_info {
  int32 r_address;
  uint32 r_info;
};

/* standard relocs: M68k, x86, ... */
#define RSTDM_symbolnum 0xffffff00  /* ordinal number of add symbol */
#define RSTDS_symbolnum 8
#define RSTDM_pcrel     0x00000080  /* 1 if value should be pc-relative */
#define RSTDS_pcrel     7
#define RSTDM_length    0x00000060  /* log base 2 of value's width */
#define RSTDS_length    5
#define RSTDM_extern    0x00000010  /* 1 if need to add symbol to value */
#define RSTDS_extern    4
#define RSTDM_baserel   0x00000008  /* linkage table relative */
#define RSTDS_baserel   3
#define RSTDM_jmptable  0x00000004  /* relocate to jump table */
#define RSTDS_jmptable  2
#define RSTDM_relative  0x00000002  /* load address relative */
#define RSTDS_relative  1
#define RSTDM_copy      0x00000001  /* run time copy */
#define RSTDS_copy      0


/* Symbol table entry format */
struct nlist32 {
  union {
    const char *n_name;     /* symbol name (in memory) */
    int32 n_strx;           /* file string table offset (on disk) */
  } n_un;
  uint8 n_type;             /* type defines */
  int8 n_other;             /* spare */
  int16 n_desc;             /* used by stab entries */
  uint32 n_value;           /* address/value of the symbol */
};

#define N_UNDF  0x00        /* undefined */
#define N_ABS   0x02        /* absolute address */
#define N_TEXT  0x04        /* text segment */
#define N_DATA  0x06        /* data segment */
#define N_BSS   0x08        /* bss segment */
#define N_INDR  0x0a        /* alias definition */
#define N_SIZE  0x0c        /* pseudo type, defines a symbol's size */
#define N_COMM  0x12        /* common reference */
#define N_SETA  0x14        /* absolute set element symbol */
#define N_SETT  0x16        /* text set element symbol */
#define N_SETD  0x18        /* data set element symbol */
#define N_SETB  0x1a        /* bss set element symbol */
#define N_SETV  0x1c        /* set vector symbol */
#define N_FN    0x1e        /* file name (N_EXT on) */
#define N_WARN  0x1e        /* warning message (N_EXT off) */

#define N_EXT   0x01        /* external (global) bit, OR'ed in */
#define N_TYPE  0x1e        /* mask for all the type bits */
#define N_STAB  0x0e0       /* mask for debugger symbols */

/* n_other & 0x0f */
#define AUX_OBJECT      1
#define AUX_FUNC        2
#define AUX_LABEL       3
/* n_other & 0xf0 >> 4 */
#define BIND_LOCAL      0   /* not used? */
#define BIND_GLOBAL     1   /* not used? */
#define BIND_WEAK       2
