/* $VER: vlink t_aout.c V0.8 (04.02.01)
 *
 * This file is part of vlink, a portable linker for multiple
 * object formats.
 * Copyright (c) 1997-2001  Frank Wille
 *
 * vlink is freeware and part of the portable and retargetable ANSI C
 * compiler vbcc, copyright (c) 1995-2001 by Volker Barthelmann.
 * vlink may be freely redistributed as long as no modifications are
 * made and nothing is charged for it. Non-commercial usage is allowed
 * without any restrictions.
 * EVERY PRODUCT OR PROGRAM DERIVED DIRECTLY FROM MY SOURCE MAY NOT BE
 * SOLD COMMERCIALLY WITHOUT PERMISSION FROM THE AUTHOR.
 *
 * v0.8  (19.01.01) phx
 *       File created.
 */

#if defined(AOUT_BSDM68K) || defined(AOUT_BSDM68K4K) || \
    defined(AOUT_SUN010) || defined(AOUT_SUN020)
#define T_AOUT_C
#include "vlink.h"
#include "aout.h"

#define TEXTNAME _text
#define DATANAME _data
#define BSSNAME  _bss

static uint8 aout_cmpsecflags(uint8,uint8);
static struct Section *aout_bssdefault(struct ObjectUnit *);
static int aout_targetlink(struct GlobalVars *,struct LinkedSection *,
                           struct Section *);
static struct Symbol *aout_lnksym(struct GlobalVars *,struct Section *,
                                  struct XReference *,struct FFFuncs *);
static void aout_setlnksym(struct GlobalVars *,struct Symbol *,
                           struct XReference *,struct FFFuncs *);

static void aoutstd_readconv(struct GlobalVars *,struct LinkFile *);


#ifdef AOUT_BSDM68K
static int aoutbsd68k_identify(char *,uint8 *,unsigned long);
static void aoutbsd68k_writeshared(struct GlobalVars *,FILE *);
static void aoutbsd68k_writeobject(struct GlobalVars *,FILE *);
static void aoutbsd68k_writeexec(struct GlobalVars *,FILE *);

struct FFFuncs fff_aoutbsd68k = {
  "aoutbsd68k",
  NULL,
  NULL,
  sizeof(struct aout_hdr),
  aoutbsd68k_identify,
  aoutstd_readconv,
  aout_cmpsecflags,
  aout_bssdefault,
  aout_targetlink,
  aout_lnksym,
  aout_setlnksym,
  aoutbsd68k_writeobject,
  aoutbsd68k_writeshared,
  aoutbsd68k_writeexec,
  0x2000,
  4,
  0x7ff0, /* @@@ ? */
  NULL,
  MID_M68K,
  0,
  1 /* big endian */
};
#endif

#ifdef AOUT_BSDM68K4K
static int aoutbsd68k4k_identify(char *,uint8 *,unsigned long);

struct FFFuncs fff_aoutbsd68k4k = {
  "aoutbsd68k4k",
  NULL,
  NULL,
  sizeof(struct aout_hdr),
  aoutbsd68k4k_identify,
  aoutstd_readconv,
  aout_cmpsecflags,
  aout_bssdefault,
  aout_targetlink,
  aout_lnksym,
  aout_setlnksym,
  aoutbsd68k_writeobject,
  aoutbsd68k_writeshared,
  aoutbsd68k_writeexec,
  0x1000,
  4,
  0x7ff0, /* @@@ ? */
  NULL,
  MID_M68K4K,
  0,
  1 /* big endian */
};
#endif

#ifdef AOUT_SUN010
static int aoutsun010_identify(char *,uint8 *,unsigned long);
static void aoutsun010_writeshared(struct GlobalVars *,FILE *);
static void aoutsun010_writeobject(struct GlobalVars *,FILE *);
static void aoutsun010_writeexec(struct GlobalVars *,FILE *);

struct FFFuncs fff_aoutsun010 = {
  "aoutsun010",
  NULL,
  NULL,
  sizeof(struct aout_hdr),
  aoutsun010_identify,
  aoutstd_readconv,
  aout_cmpsecflags,
  aout_bssdefault,
  aout_targetlink,
  aout_lnksym,
  aout_setlnksym,
  aoutsun010_writeobject,
  aoutsun010_writeshared,
  aoutsun010_writeexec,
  0x2000, /* @@@ ? */
  4,
  0x7ff0, /* @@@ ? */
  NULL,
  MID_SUN010,
  0,
  1 /* big endian */
};
#endif

#ifdef AOUT_SUN020
static int aoutsun020_identify(char *,uint8 *,unsigned long);
static void aoutsun020_writeshared(struct GlobalVars *,FILE *);
static void aoutsun020_writeobject(struct GlobalVars *,FILE *);
static void aoutsun020_writeexec(struct GlobalVars *,FILE *);

struct FFFuncs fff_aoutsun020 = {
  "aoutsun020",
  NULL,
  NULL,
  sizeof(struct aout_hdr),
  aoutsun020_identify,
  aoutstd_readconv,
  aout_cmpsecflags,
  aout_bssdefault,
  aout_targetlink,
  aout_lnksym,
  aout_setlnksym,
  aoutsun020_writeobject,
  aoutsun020_writeshared,
  aoutsun020_writeexec,
  0x2000, /* @@@ ? */
  4,
  0x7ff0, /* @@@ ? */
  NULL,
  MID_SUN020,
  0,
  1 /* big endian */
};
#endif


/* common a.out linker symbols */
static char *aout_symnames[] = {
  "_etext","_edata","_end",
  "__GLOBAL_OFFSET_TABLE_",
  "__PROCEDURE_LINKAGE_TABLE_",
  "__DYNAMIC"
};
#define AOUTSTD_LNKSYMS 6
#define ETEXTSYM        0
#define EDATASYM        1
#define ENDSYM          2
#define GOTSYM          3
#define PLTSYM          4
#define DYNAMICSYM      5

/* a.out default sections names */
static char *_text = ".text";
static char *_data = ".data";
static char *_bss =  ".bss";

static struct ar_info ai;  /* for scanning library archives */



/*****************************************************************/
/*                          Read a.out                           */
/*****************************************************************/


static uint32 aout_rd32(uint32 *p,int be)
{
  return (be ? read32be(p) : read32le(p));
}

static uint32 aout_txtoffset(struct aout_hdr *hdr,int be)
{
  return (sizeof(struct aout_hdr));
}

static uint32 aout_datoffset(struct aout_hdr *hdr,int be)
{
  return ((GETMAGIC(hdr) == ZMAGIC) ?
          (aout_rd32(&hdr->a_text,be)) :
          (sizeof(struct aout_hdr) + aout_rd32(&hdr->a_text,be)));
}

static uint32 aout_bssoffset(struct aout_hdr *hdr,int be)
{
  return (aout_datoffset(hdr,be) + aout_rd32(&hdr->a_data,be));
}

static uint32 aout_treloffset(struct aout_hdr *hdr,int be)
{
  return (aout_bssoffset(hdr,be));
}

static uint32 aout_dreloffset(struct aout_hdr *hdr,int be)
{
  return (aout_treloffset(hdr,be) + aout_rd32(&hdr->a_trsize,be));
}

static uint32 aout_symoffset(struct aout_hdr *hdr,int be)
{
  return (aout_dreloffset(hdr,be) + aout_rd32(&hdr->a_drsize,be));
}

static uint32 aout_stroffset(struct aout_hdr *hdr,int be)
{
  return (aout_symoffset(hdr,be) + aout_rd32(&hdr->a_syms,be));
}


static int aout_identify(struct FFFuncs *ff,char *name,struct aout_hdr *p,
                         unsigned long plen,uint32 mid)
/* check a possible a.out file against the requirements, then */
/* return its type (object, library, shared object) */
{
  bool arflag = FALSE;

  if (plen < ff->headersize)
    return (ID_UNKNOWN);

  if (ar_init(&ai,(char *)p,plen,name)) {
    /* library archive detected, extract 1st archive member */
    arflag = TRUE;
    if (!(ar_extract(&ai))) {
      error(38,name);  /* Empty archive ignored */
      return (ID_LIBARCH);
    }
    p = (struct aout_hdr *)ai.data;
  }

  if (GETMID(p)==mid || GETMID(p)==0) {
    uint32 mag = GETMAGIC(p);
    int fl = GETFLAGS(p) & EX_DPMASK;

    if (mag==OMAGIC || mag==NMAGIC || mag==ZMAGIC || mag==QMAGIC) {
      /* valid a.out format for this machine detected! */
      switch (mag) {
        case OMAGIC:
          return (arflag ? ID_LIBARCH : ID_OBJECT);
        case NMAGIC:
        case ZMAGIC:
          if (arflag) {
            if (fl == EX_DYNAMIC|EX_PIC)
              error(39,name,ff->tname); /* no shared objects in lib archives */
            else
              error(40,name,ff->tname); /* no executables in lib archives */
          }
          if (fl == EX_DYNAMIC|EX_PIC)
            return (ID_SHAREDOBJ);
          return (ID_EXECUTABLE);
        case QMAGIC:
          error(84,name); /* QMAGIC is deprecated */
          break;
      }
    }
  }
  return (ID_UNKNOWN);
}


static void aout_check_ar_type(struct FFFuncs *ff,char *name,
                               struct aout_hdr *hdr)
/* check all library archive members before conversion */
{
  if (GETMID(hdr)==ff->id || GETMID(hdr)==0) {
    switch (GETMAGIC(hdr)) {
      case NMAGIC:
      case ZMAGIC:
        if (GETFLAGS(hdr)&EX_DPMASK == (EX_DYNAMIC|EX_PIC))
          error(39,name,ff->tname); /* no shared objects in lib archives */
        else
          error(40,name,ff->tname); /* no executables in lib archives */
        return;
      case QMAGIC:
        error(84,name); /* QMAGIC is deprecated */
      case OMAGIC:
        return;
      default:
        break;
    }
  }
  error(41,name,ff->tname);
}


static struct Section *aout_allocsection(struct ObjectUnit *ou,
                                         const char *sec_name,
                                         uint8 *data,unsigned long size,
                                         uint8 type,uint8 flags,
                                         uint8 protection,uint8 alignment)
{
  struct Section *s = create_section(ou,(char *)sec_name,data,size);

  s->type = type;
  s->flags = flags;
  s->protection = protection;
  s->alignment = alignment;
  return (s);
}


static struct Section *aout_bssdefault(struct ObjectUnit *ou)
{
  return (aout_allocsection(ou,BSSNAME,NULL,0,ST_UDATA,
                            SF_UNINITIALIZED,SP_READ|SP_WRITE,
                            fff[ou->lnkfile->format]->minalignment));
}


static void aout_create_section(struct ObjectUnit *ou,struct FFFuncs *ff,
                                const char *sec_name,uint8 *data,
                                unsigned long size,uint8 type)
{
  struct LinkFile *lf = ou->lnkfile;
  struct Section *s;
  uint8 flags=0,protection;

  if (type != ST_UDATA) {
    if (data+size > lf->data+lf->length)  /* illegal section offset */
      error(49,lf->pathname,sec_name,lf->objname);
  }
  s = create_section(ou,(char *)sec_name,data,size);
  switch (type) {
    case ST_CODE:
      protection = SP_READ|SP_EXEC;
      break;
    case ST_DATA:
      protection = SP_READ|SP_WRITE;
      break;
    case ST_UDATA:
      protection = SP_READ|SP_WRITE;
      flags = SF_UNINITIALIZED;
      break;
  }
  s = aout_allocsection(ou,sec_name,data,size,type,flags,protection,
                        ff->minalignment);
  addtail(&ou->sections,&s->n);  /* add Section to ObjectUnit */
}


static void aout_make_sections(struct ObjectUnit *ou,struct aout_hdr *hdr,
                               int be)
/* creates up to three sections from an a.out file */
{
  struct FFFuncs *ff = fff[ou->lnkfile->format];
  uint32 size;

  if (size = aout_rd32(&hdr->a_text,be))
    aout_create_section(ou,ff,TEXTNAME,((uint8 *)hdr)+aout_txtoffset(hdr,be),
                        size,ST_CODE);

  if (size = aout_rd32(&hdr->a_data,be))
    aout_create_section(ou,ff,DATANAME,((uint8 *)hdr)+aout_datoffset(hdr,be),
                        size,ST_DATA);

  if (size = aout_rd32(&hdr->a_bss,be))
    aout_create_section(ou,ff,BSSNAME,((uint8 *)hdr)+aout_bssoffset(hdr,be),
                        size,ST_UDATA);
}


static void check_strtab(struct LinkFile *lf,struct aout_hdr *hdr,int be)
{
  char *strtab = ((char *)hdr) + aout_stroffset(hdr,be);

  if (((uint8 *)strtab < lf->data) || ((uint8 *)strtab +
       aout_rd32((uint32 *)strtab,be) > (lf->data + lf->length)))
    error(85,lf->pathname,"string",lf->objname);  /* illegal offset */
}


static char *get_symname(struct LinkFile *lf,struct aout_hdr *hdr,int be,
                         int32 offset)
{
  char *symname = ((char *)hdr) + aout_stroffset(hdr,be) + offset;

  if (((uint8 *)symname < lf->data) ||
      ((uint8 *)symname > (lf->data + lf->length)))  /* illegal offset? */
    error(87,lf->pathname,offset,lf->objname);

  return (symname);
}


static void aout_symbols(struct GlobalVars *gv,struct ObjectUnit *ou,
                         struct aout_hdr *hdr,int be)
/* reads all symbols from an a.out and converts them into internal fmt. */
{
  struct LinkFile *lf = ou->lnkfile;
  struct FFFuncs *ff = fff[lf->format];
  struct nlist32 *nlst = (struct nlist32 *)(((char *)hdr) +
                                            aout_symoffset(hdr,be));
  struct Section *firstsec;
  uint32 symtabsize = aout_rd32(&hdr->a_syms,be);
  uint32 txtaddr = (GETMAGIC(hdr) == ZMAGIC) ? sizeof(struct aout_hdr) : 0;
  uint32 dataddr = aout_rd32(&hdr->a_text,be);
  uint32 bssaddr = dataddr + aout_rd32(&hdr->a_data,be);
  int i;

  if (((uint8 *)nlst < lf->data) || 
      ((uint8 *)nlst + symtabsize > lf->data + lf->length))
    error(85,lf->pathname,"symbol",lf->objname);  /* illegal offset */

  if (symtabsize % sizeof(struct nlist32) != 0)  /* symtab sanity check */
    error(86,lf->pathname,"symbol",lf->objname,sizeof(struct nlist32));

  check_strtab(lf,hdr,be);  /* strtab sanity check */

  if (listempty(&ou->sections))  /* not a single section? */
    firstsec = ff->bssdefault(ou);
  else
    firstsec = (struct Section *)ou->sections.first;

  /* read the symbols */
  for (i=0; i<(int)(symtabsize/sizeof(struct nlist32)); i++,nlst++) {
    char *symname,*indirname = NULL;
    uint8 type=SYM_RELOC,objinfo,objbind;
    uint32 saddr=0,size=0;
    uint32 value=aout_rd32(&nlst->n_value,be);
    struct Section *sec = NULL;

    symname = get_symname(lf,hdr,be,aout_rd32(&nlst->n_un.n_strx,be));

    switch (nlst->n_other & 0x0f) {
      case AUX_FUNC:
      case AUX_LABEL: /* @@@ what's label? */
        objinfo = SYMI_FUNC;
        break;
      case AUX_OBJECT:
        objinfo = SYMI_OBJECT;
        break;
      default:
        objinfo = SYMI_NOTYPE;
        break;
    }

    objbind = (nlst->n_type & N_EXT) ? SYMB_GLOBAL : SYMB_LOCAL;
    switch ((nlst->n_other & 0xf0) >> 4) {
      /* seems that local and global are not used in a.out... */
      case BIND_LOCAL:
        /*objbind = SYMB_LOCAL;*/
        break;
      case BIND_GLOBAL:
        /*objbind = SYMB_GLOBAL;*/
        break;
      case BIND_WEAK:
        objbind = SYMB_WEAK;
        break;
      default:  /* illegal binding type */
        error(88,lf->pathname,symname,(nlst->n_other&0xf0)>>4,lf->objname);
        break;
    }

    /* @@@ Debugging symbols are currently ignored... */
    if (nlst->n_type & N_STAB)
      continue;

    switch (nlst->n_type & N_TYPE) {
      case N_UNDF:
        if (value) {
          /* undefd. symbol with a value is assumed to be a common symbol */
          sec = firstsec;
          type = SYM_COMMON;
          size = value;
          value = ff->minalignment;
        }
        /* ignore undefined symbols for now */
        break;

      /* assign a section for ABS and COMMON symbols, to prevent */
      /* accidental NULL-pointer references */
      case N_ABS:
        sec = firstsec;
        type = SYM_ABS;
        break;

#if 0 /* @@@ N_UNDF is used instead? */
      case N_COMM:
        sec = firstsec;
        type = SYM_COMMON;
        size = value;
        break;
#endif

      case N_TEXT:
        if (!(sec = find_sect_type(ou,ST_CODE,SP_READ|SP_EXEC)))
          sec = aout_allocsection(ou,TEXTNAME,NULL,0,ST_CODE,
                                  0,SP_READ|SP_EXEC,ff->minalignment);
        saddr = txtaddr;
        break;

      case N_DATA:
        if (!(sec = find_sect_type(ou,ST_DATA,SP_READ|SP_WRITE)))
          sec = aout_allocsection(ou,DATANAME,NULL,0,ST_DATA,
                                  0,SP_READ|SP_WRITE,ff->minalignment);
        saddr = dataddr;
        break;

      case N_BSS:
        if (!(sec = find_sect_type(ou,ST_UDATA,SP_READ|SP_WRITE)))
          sec = aout_allocsection(ou,BSSNAME,NULL,0,ST_UDATA,
                                  SF_UNINITIALIZED,SP_READ|SP_WRITE,
                                  ff->minalignment);
        saddr = bssaddr;
        break;

      case N_INDR:
        sec = firstsec;
        type = SYM_INDIR;
        indirname = get_symname(lf,hdr,be,
                                aout_rd32(&(nlst+1)->n_un.n_strx,be));
        break;

      case N_SIZE:
        /* ignored */
        error(96,lf->pathname,(int)value,lf->objname);
        break;

      case N_FN:
        if (nlst->n_type & N_EXT) {
          sec = firstsec;
          type = SYM_ABS;
          objinfo = SYMI_FILE;
          objbind = SYMB_LOCAL;
          break;
        }

      default:
        ierror("aout_symbols(): Symbol %s in %s has type %d, "
               "which is currently not supported",
               symname,lf->pathname,nlst->n_type&N_TYPE);
        break;
    }

    if (sec) {
      struct Symbol *sym;

      /* check for symbol size: following symbol has to be type = N_SIZE */
      /* and the same symbol name */
      if ((i+1)<(int)(symtabsize/sizeof(struct nlist32)) &&
          ((nlst+1)->n_type & N_TYPE) == N_SIZE) {
        if (!strcmp(symname,get_symname(lf,hdr,be,
                    aout_rd32(&(nlst+1)->n_un.n_strx,be)))) {
          i++;  /* skip symbol */
          nlst++;
          size = aout_rd32(&nlst->n_value,be);
        }
      }

      /* convert a.out specific file offset into a section offset */
      value -= saddr;

      /* add a new symbol definition */
      if (objbind == SYMB_LOCAL) {
        /* always define local symbols - multiple definitions allowed */
        addlocsymbol(gv,sec,symname,indirname,value,type,0,objinfo,size);
      }
      else {
        if (sym = addsymbol(gv,sec,symname,indirname,value,type,0,objinfo,
                            objbind,size)) {
          /* symbol is multiply defined! */
          struct LinkFile *lf2 = sym->relsect->obj->lnkfile;
          char buf[256];  /* @@@ dangerous... */

          if (lf2->type == ID_LIBARCH)
            sprintf(buf,"%s (%s)",lf2->filename,lf2->objname);
          else
            strcpy(buf,lf2->objname);
          error(56,lf->pathname,symname,lf->objname,buf);
        }
      }
#if 0 /* @@@ to be implemented... */
      if (objbind == SYMB_GLOBAL && !gv->dest_object) {
      /* symbol describes an automatic vbcc con-/destructor function ? */
        if (!strncmp(vbcc_INIT,symname,sizeof(vbcc_INIT)-1))
          new_priptr(sec->obj,xtors_objname,elf32_symnames[CTORS],
                     get_xtors_pri(symname+(sizeof(vbcc_INIT)-1)),symname,0);
        else if (!strncmp(vbcc_EXIT,symname,sizeof(vbcc_EXIT)-1))
          new_priptr(sec->obj,xtors_objname,elf32_symnames[DTORS],
                     get_xtors_pri(symname+(sizeof(vbcc_EXIT)-1)),symname,0);
      }
#endif
    }
  }
}


static void aout_newreloc(struct GlobalVars *gv,struct aout_hdr *hdr,
                          struct Section *sec,uint32 symidx,
                          bool xtern,uint8 rtype,
                          uint32 offs,int be)
{
  const char *fn = "aout_newreloc(): ";
  struct ObjectUnit *ou = sec->obj;
  struct LinkFile *lf = ou->lnkfile;
  int32 a; /* addend */

  if (xtern) {
    /* relocation by an external reference to an unknown symbol */
    char *strtab = ((char *)hdr) + aout_stroffset(hdr,be);
    struct nlist32 *nlst = (struct nlist32 *)(((char *)hdr) +
                                              aout_symoffset(hdr,be));
    char *xrefname = strtab + aout_rd32(&nlst[symidx].n_un.n_strx,be);

    switch (rtype & R_MASK) {
      case R_ADDR:
      case R_BASEREL:
        a = (int32)readsection(gv,sec->data+offs,rtype);
        break;
      case R_REL:
        a = (int32)offs + (int32)readsection(gv,sec->data+offs,rtype);
        break;
      default:
        /* @@@ other reloc types still missing */
        ierror("%sUnknown reloc type %d",fn,(int)rtype);
        break;
    }

    if (nlst[symidx].n_type == N_EXT|N_UNDF) {
      addxref(gv,sec,xrefname,offs,rtype,relocsize(rtype),a);
    }
    else
      error(91,lf->pathname,xrefname,sec->name);  /* illegal ext. ref. */
  }

  else {
    /* local relocation */
    struct Section *rsec;
    uint32 saddr; /* section base address in object file (.text = 0) */

    if (symidx & ~N_TYPE)  /* illegal nlist type in relocation */
      error(92,lf->pathname,symidx,sec->name,lf->objname,offs);

    switch (symidx & N_TYPE) {
      case N_TEXT:
        if (!(rsec = find_sect_type(ou,ST_CODE,SP_READ|SP_EXEC)))
          ierror("%sno .text for reloc found",fn);
        saddr = 0;
        break;

      case N_DATA:
        if (!(rsec = find_sect_type(ou,ST_DATA,SP_READ|SP_WRITE)))
          ierror("%sno .data for reloc found",fn);
        saddr = aout_rd32(&hdr->a_text,be);
        break;

      case N_BSS:
        if (!(rsec = find_sect_type(ou,ST_UDATA,SP_READ|SP_WRITE)))
          ierror("%sno .bss for reloc found",fn);
        saddr = aout_rd32(&hdr->a_text,be) + aout_rd32(&hdr->a_data,be);
        break;

      default:
        ierror("%slocal reloc with nlist type %lu is not supported",
               fn,symidx&N_TYPE);
        break;
    }

    /* read addend from section */
    switch (rtype & R_MASK) {
      case R_ADDR:
        a = (int32)(readsection(gv,sec->data+offs,rtype) - saddr);
        break;
      case R_REL: /* @@@ do they exist? */
        a = (int32)offs + (int32)readsection(gv,sec->data+offs,rtype);
        break;
      case R_BASEREL:
        a = (int32)readsection(gv,sec->data+offs,rtype);
        break;
      default:
        /* @@@ other reloc types still missing */
        ierror("%sUnknown local reloc type %d",fn,(int)rtype);
        break;
    }

    addreloc(sec,rsec,0,offs,rtype,a);
  }
}


static uint8 aout_cmpsecflags(uint8 oldflags,uint8 newflags)
/* compare and verify target-specific section flags, */
/* return 0xff if sections are incompatible, otherwise return new flags */
{
  return (oldflags|newflags);
}


static int aout_targetlink(struct GlobalVars *gv,struct LinkedSection *ls,
                           struct Section *s)
/* returns 1, if target requires the combination of the two sections, */
/* returns -1, if target don't want to combine them, */
/* returns 0, if target doesn't care - standard linking rules are used. */
{
  return (0);
}


static struct Symbol *aout_lnksym(struct GlobalVars *gv,
                                  struct Section *sec,
                                  struct XReference *xref,
                                  struct FFFuncs *tf)
/* Check for common a.out linker symbols. */
{
  struct Symbol *sym;
  int i;

  if (sym = findlnksymbol(tf,xref->name))
    return (sym);

  if (!gv->dest_object) {
    for (i=0; i<AOUTSTD_LNKSYMS; i++) {
      if (!strcmp(aout_symnames[i],xref->name)) {
        sym = addlnksymbol(tf,NULL,aout_symnames[i],0,SYM_ABS,
                           SYMF_LNKSYM,SYMI_OBJECT,SYMB_GLOBAL,0);
        sym->extra = i; /* for easy ident. in aout_setlnksym */
        switch (i) {
          case ETEXTSYM:
          case EDATASYM:
          case ENDSYM:
          case GOTSYM:
          case PLTSYM:
          case DYNAMICSYM:
            sym->type = SYM_RELOC;
            break;
        }
        return (sym);  /* new linker symbol created */
      }
    }
  }
  return (NULL);
}


static void aout_setlnksym(struct GlobalVars *gv,struct Symbol *xdef,
                           struct XReference *xref,struct FFFuncs *tf)
/* Initialize common a.out linker symbol structure during resolve_xref() */
{
  if (xdef->flags & SYMF_LNKSYM) {
    struct LinkedSection *ls;

    switch (xdef->extra) {
      case ETEXTSYM:
        if (ls = find_lnksec(gv,TEXTNAME,ST_CODE,0,0)) {
          xdef->value = ls->size;
        }
        else {
          if (!(ls = find_lnksec(gv,DATANAME,ST_DATA,0,0)))
            if (!(ls = find_lnksec(gv,BSSNAME,ST_UDATA,0,0)))
              break;
        }
        xdef->relsect = (struct Section *)ls->sections.first;
        break;

      case EDATASYM:
        if (ls = find_lnksec(gv,DATANAME,ST_DATA,0,0))
          xdef->value = ls->size;
        else if (ls = find_lnksec(gv,TEXTNAME,ST_CODE,0,0))
          xdef->value = ls->size;
        else if (!(ls = find_lnksec(gv,BSSNAME,ST_UDATA,0,0)))
          break;
        xdef->relsect = (struct Section *)ls->sections.first;
        break;

      case ENDSYM:
        if (ls = find_lnksec(gv,BSSNAME,ST_UDATA,0,0))
          xdef->value = ls->size;
        else if (ls = find_lnksec(gv,DATANAME,ST_DATA,0,0))
          xdef->value = ls->size;
        else if (ls = find_lnksec(gv,TEXTNAME,ST_CODE,0,0))
          xdef->value = ls->size;
        else
          break;
        xdef->relsect = (struct Section *)ls->sections.first;
        break;

      /*@@@ GOTSYM, PLTSYM, DYNAMICSYM */
    }
    xdef->flags &= ~SYMF_LNKSYM;  /* do not init again */
  }
}


static void aoutstd_relocs(struct GlobalVars *gv,struct ObjectUnit *ou,
                           struct aout_hdr *hdr,int be,
                           struct relocation_info *reloc,uint32 rsize,
                           struct Section *sec)
/* reads all relocations in standard format for a section */
{
  struct LinkFile *lf = ou->lnkfile;

  if (sec) {
    int i;

    if (((uint8 *)reloc < lf->data) ||
        (((uint8 *)reloc)+rsize > lf->data + lf->length))
      error(85,lf->pathname,"stdreloc",lf->objname);  /* illegal offset */

    if (rsize % sizeof(struct relocation_info) != 0)  /* reloc sanity check */
      error(86,lf->pathname,"stdreloc",lf->objname,
            sizeof(struct relocation_info));

    for (i=0; i<(rsize/sizeof(struct relocation_info)); i++,reloc++) {
      static uint8 raddr[4] = { R_ADDR8,R_ADDR16,R_ADDR32,0 };
      static uint8 rrelpc[4] = { R_REL8,R_REL16,R_REL32,0 };
      static uint8 rbrel[4] = { R_BASEREL8,R_BASEREL16,R_BASEREL32,0 };
      uint32 info = aout_rd32(&reloc->r_info,be);
      uint32 symnum = (info & RSTDM_symbolnum) >> RSTDS_symbolnum;
      int size = (info & RSTDM_length) >> RSTDS_length;
      uint8 rtype;

      /* @@@ only a few relocations supported for now */
      if ((info & (RSTDM_pcrel | RSTDM_baserel | RSTDM_jmptable)) == 0) {
        if (info & (RSTDM_copy | RSTDM_relative)) {
          /* @@@ COPY- and REL-relocations are treated as normal ones! */
          error(95,lf->pathname,(int)(info & RSTDM_pcrel) >> RSTDS_pcrel,
                (int)(info & RSTDM_length) >> RSTDS_length,
                (int)(info & RSTDM_extern) >> RSTDS_extern,
                (int)(info & RSTDM_baserel) >> RSTDS_baserel,
                (int)(info & RSTDM_jmptable) >> RSTDS_jmptable,
                (int)(info & RSTDM_relative) >> RSTDS_relative,
                (int)(info & RSTDM_copy) >> RSTDS_copy,sec->name,
                lf->objname,(uint32)aout_rd32(&reloc->r_address,be));
        }
        rtype = raddr[size];  /* R_ADDRx absolute address relocation */
      }
      else if ((info & RSTDM_baserel) && (info & (RSTDM_pcrel |
               RSTDM_jmptable | RSTDM_relative | RSTDM_copy)) == 0) {
        rtype = rbrel[size];  /* R_BASERELx base relative relocation */
      }
      else if ((info & RSTDM_pcrel) && (info & (RSTDM_baserel |
               RSTDM_jmptable | RSTDM_relative | RSTDM_copy)) == 0) {
        rtype = rrelpc[size]; /* R_RELx PC relative relocation */
      }
      else
        ierror("aoutstd_relocs(): %s (%s): Reloc type "
               "<pcrel=%d len=%d extern=%d baserel=%d jmptab=%d "
               "rel=%d copy=%d> in %s is currently not supported",
               lf->pathname,lf->objname,
               (int)(info & RSTDM_pcrel) >> RSTDS_pcrel,
               (int)(info & RSTDM_length) >> RSTDS_length,
               (int)(info & RSTDM_extern) >> RSTDS_extern,
               (int)(info & RSTDM_baserel) >> RSTDS_baserel,
               (int)(info & RSTDM_jmptable) >> RSTDS_jmptable,
               (int)(info & RSTDM_relative) >> RSTDS_relative,
               (int)(info & RSTDM_copy) >> RSTDS_copy,
               sec->name);

      if (rtype) {
        /* create a.out relocation */
        aout_newreloc(gv,hdr,sec,symnum,(info & RSTDM_extern)!=0,
                      rtype,aout_rd32(&reloc->r_address,be),be);
      }
      else {
        /* illegal a.out relocation */
        error(90,lf->pathname,sec->name,lf->objname,
              (uint32)aout_rd32(&reloc->r_address,be),
              (int)(info & RSTDM_pcrel) >> RSTDS_pcrel,
              (int)(info & RSTDM_length) >> RSTDS_length,
              (int)(info & RSTDM_extern) >> RSTDS_extern,
              (int)(info & RSTDM_baserel) >> RSTDS_baserel,
              (int)(info & RSTDM_jmptable) >> RSTDS_jmptable,
              (int)(info & RSTDM_relative) >> RSTDS_relative,
              (int)(info & RSTDM_copy) >> RSTDS_copy);
      }
    }
  }

  else if (rsize) {
    /* no section for these relocations! */
    error(89,lf->pathname,lf->objname);
  }
}


static void aoutstd_read(struct GlobalVars *gv,struct LinkFile *lf,
                         struct aout_hdr *hdr)
/* read an a.out file with standard relocations */
{
  int be = fff[lf->format]->big_endian != 0;
  int so = (GETFLAGS(hdr)&EX_DPMASK) == (EX_DYNAMIC|EX_PIC);
  struct ObjectUnit *u;

  if (lf->type == ID_LIBARCH)  /* check ar-member for correct format */
    aout_check_ar_type(fff[lf->format],lf->pathname,hdr);

  u = create_objunit(lf,lf->objname);

  aout_make_sections(u,hdr,be);   /* create up to 3 sections */
  aout_symbols(gv,u,hdr,be);      /* read all symbols */

  aoutstd_relocs(gv,u,hdr,be,(struct relocation_info *)    /* .text relocs */
                 (((uint8 *)hdr)+aout_treloffset(hdr,be)),
                 aout_rd32(&hdr->a_trsize,be),
                 find_sect_type(u,ST_CODE,SP_READ|SP_EXEC));
  aoutstd_relocs(gv,u,hdr,be,(struct relocation_info *)    /* .data relocs */
                 (((uint8 *)hdr)+aout_dreloffset(hdr,be)),
                 aout_rd32(&hdr->a_drsize,be),
                 find_sect_type(u,ST_DATA,SP_READ|SP_WRITE));

  /* add new object unit to the appropriate list */
  add_objunit(gv,u,FALSE);
}


static void aoutstd_readconv(struct GlobalVars *gv,struct LinkFile *lf)
/* Read a.out executable / object / shared obj. with standard relocs */
{
  if (lf->type == ID_LIBARCH) {
    if (ar_init(&ai,(char *)lf->data,lf->length,lf->filename)) {
      while (ar_extract(&ai)) {
        lf->objname = allocstring(ai.name);
        aoutstd_read(gv,lf,(struct aout_hdr *)ai.data);
      }
    }
    else
      ierror("aoutstd_readconv(): archive %s corrupted since last access",
             lf->pathname);
  }
  else {
    lf->objname = lf->filename;
    aoutstd_read(gv,lf,(struct aout_hdr *)lf->data);
  }
}


#ifdef AOUT_BSDM68K
static int aoutbsd68k_identify(char *name,uint8 *p,unsigned long plen)
{
  return (aout_identify(&fff_aoutbsd68k,name,(struct aout_hdr *)p,
                        plen,MID_M68K));
}
#endif


#ifdef AOUT_BSDM68K4K
static int aoutbsd68k4k_identify(char *name,uint8 *p,unsigned long plen)
{
  return (aout_identify(&fff_aoutbsd68k4k,name,(struct aout_hdr *)p,
                        plen,MID_M68K4K));
}
#endif


#ifdef AOUT_SUN010
static int aoutsun010_identify(char *name,uint8 *p,unsigned long plen)
{
  return (aout_identify(&fff_aoutsun010,name,(struct aout_hdr *)p,
                        plen,MID_SUN010));
}
#endif


#ifdef AOUT_SUN020
static int aoutsun020_identify(char *name,uint8 *p,unsigned long plen)
{
  return (aout_identify(&fff_aoutsun020,name,(struct aout_hdr *)p,
                        plen,MID_SUN020));
}
#endif



/*****************************************************************/
/*                          Write a.out                          */
/*****************************************************************/


#if defined(AOUT_BSDM68K) || defined(AOUT_BSDM68K4K)
static void aoutbsd68k_writeshared(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutbsd68k shared object (which is pos. independant) */
{
  ierror("aoutbsd68k_writeshared(): Shared object generation has not "
         "yet been implemented");
}


static void aoutbsd68k_writeobject(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutbsd68k relocatable object file */
{
  ierror("aoutbsd68k_writeexec(): Object file generation has not "
         "yet been implemented");
}


static void aoutbsd68k_writeexec(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutbsd68k executable file (with absolute addresses) */
{
  ierror("aoutbsd68k_writeexec(): Executable file generation has not "
         "yet been implemented");
}
#endif


#ifdef AOUT_SUN010
static void aoutsun010_writeshared(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutsun010 shared object (which is pos. independant) */
{
  ierror("aoutsun010_writeshared(): Shared object generation has not "
         "yet been implemented");
}


static void aoutsun010_writeobject(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutsun010 relocatable object file */
{
  ierror("aoutsun010_writeexec(): Object file generation has not "
         "yet been implemented");
}


static void aoutsun010_writeexec(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutsun010 executable file (with absolute addresses) */
{
  ierror("aoutsun010_writeexec(): Executable file generation has not "
         "yet been implemented");
}
#endif


#ifdef AOUT_SUN020
static void aoutsun020_writeshared(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutsun020 shared object (which is pos. independant) */
{
  ierror("aoutsun020_writeshared(): Shared object generation has not "
         "yet been implemented");
}


static void aoutsun020_writeobject(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutsun020 relocatable object file */
{
  ierror("aoutsun020_writeexec(): Object file generation has not "
         "yet been implemented");
}


static void aoutsun020_writeexec(struct GlobalVars *gv,FILE *f)
/* creates a target-aoutsun020 executable file (with absolute addresses) */
{
  ierror("aoutsun020_writeexec(): Executable file generation has not "
         "yet been implemented");
}
#endif


#endif
