/* $VER: vlink t_rawbin.c V0.8 (11.02.01)
 *
 * This file is part of vlink, a portable linker for multiple
 * object formats.
 * Copyright (c) 1997-2001  Frank Wille
 *
 * vlink is freeware and part of the portable and retargetable ANSI C
 * compiler vbcc, copyright (c) 1995-2001 by Volker Barthelmann.
 * vlink may be freely redistributed as long as no modifications are
 * made and nothing is charged for it. Non-commercial usage is allowed
 * without any restrictions.
 * EVERY PRODUCT OR PROGRAM DERIVED DIRECTLY FROM MY SOURCE MAY NOT BE
 * SOLD COMMERCIALLY WITHOUT PERMISSION FROM THE AUTHOR.
 *
 *
 * v0.8  (11.02.01) phx
 *       Included Motorola S-Record formats S1S9, S2S8 and S3S7.
 * v0.6e (29.08.99) phx
 *       File created.
 */


#if defined(RAWBIN1) || defined(RAWBIN2) || \
    defined(SREC19) || defined(SREC28) || defined(SREC37)
#define T_RAWBIN_C
#include "vlink.h"

#define MAXGAP 16  /* from MAXGAP bytes on, a new file will be created */
#define GAPBUFSIZE 1024
#define MAXSREC 32 /* max. number of bytes in an S1/S2/S3 record */


static int rawbin_identify(char*,uint8 *,unsigned long);
static void rawbin_readconv(struct GlobalVars *,struct LinkFile *);
static uint8 rawbin_cmpsecflags(uint8,uint8);
static struct Section *rawbin_bssdefault(struct ObjectUnit *);
static int rawbin_targetlink(struct GlobalVars *,struct LinkedSection *,
                              struct Section *);
static struct Symbol *rawbin_lnksym(struct GlobalVars *,struct Section *,
                                     struct XReference *);
static void rawbin_setlnksym(struct GlobalVars *,struct Symbol *,
                              struct XReference *);
static void rawbin_writeobject(struct GlobalVars *,FILE *);
static void rawbin_writeshared(struct GlobalVars *,FILE *);
static void rawbin_writeexec(struct GlobalVars *,FILE *,bool);
#ifdef RAWBIN1
static void rawbin_writesingle(struct GlobalVars *,FILE *);
#endif
#ifdef RAWBIN2
static void rawbin_writemultiple(struct GlobalVars *,FILE *);
#endif
#ifdef SREC19
static void srec19_write(struct GlobalVars *,FILE *);
#endif
#ifdef SREC28
static void srec28_write(struct GlobalVars *,FILE *);
#endif
#ifdef SREC37
static void srec37_write(struct GlobalVars *,FILE *);
#endif
static void calc_relocs(struct GlobalVars *,struct LinkedSection *);
static struct LinkedSection *get_next_section(struct GlobalVars *);
static void writegap(FILE *,long);


static const char defaultscript[] =
  "SECTIONS {\n"
  "  .text: { *(.text) *(CODE) *(text) }\n"
  "  .data: { *(.data) *(DATA) *(data) }\n"
  "  .bss: { *(.bss) *(BSS) *(bss) }\n"
  "}\n";


#ifdef RAWBIN1
struct FFFuncs fff_rawbin1 = {
  "rawbin1",
  defaultscript,
  NULL,
  0,
  rawbin_identify,
  rawbin_readconv,
  rawbin_cmpsecflags,
  rawbin_bssdefault,
  rawbin_targetlink,
  rawbin_lnksym,
  rawbin_setlnksym,
  rawbin_writeobject,
  rawbin_writeshared,
  rawbin_writesingle,
  0,
  2,
  0x7ff0,
  NULL,
  0,
  0,
  -1 /* endianess undefined, only write */
};
#endif

#ifdef RAWBIN2
struct FFFuncs fff_rawbin2 = {
  "rawbin2",
  defaultscript,
  NULL,
  0,
  rawbin_identify,
  rawbin_readconv,
  rawbin_cmpsecflags,
  rawbin_bssdefault,
  rawbin_targetlink,
  rawbin_lnksym,
  rawbin_setlnksym,
  rawbin_writeobject,
  rawbin_writeshared,
  rawbin_writemultiple,
  0,
  2,
  0x7ff0,
  NULL,
  0,
  0,
  -1 /* endianess undefined, only write */
};
#endif

#ifdef SREC19
struct FFFuncs fff_srec19 = {
  "srec19",
  defaultscript,
  NULL,
  0,
  rawbin_identify,
  rawbin_readconv,
  rawbin_cmpsecflags,
  rawbin_bssdefault,
  rawbin_targetlink,
  rawbin_lnksym,
  rawbin_setlnksym,
  rawbin_writeobject,
  rawbin_writeshared,
  srec19_write,
  0,
  2,
  0x7ff0,
  NULL,
  0,
  0,
  -1 /* endianess undefined, only write */
};
#endif

#ifdef SREC28
struct FFFuncs fff_srec28 = {
  "srec28",
  defaultscript,
  NULL,
  0,
  rawbin_identify,
  rawbin_readconv,
  rawbin_cmpsecflags,
  rawbin_bssdefault,
  rawbin_targetlink,
  rawbin_lnksym,
  rawbin_setlnksym,
  rawbin_writeobject,
  rawbin_writeshared,
  srec28_write,
  0,
  2,
  0x7ff0,
  NULL,
  0,
  0,
  -1 /* endianess undefined, only write */
};
#endif

#ifdef SREC37
struct FFFuncs fff_srec37 = {
  "srec37",
  defaultscript,
  NULL,
  0,
  rawbin_identify,
  rawbin_readconv,
  rawbin_cmpsecflags,
  rawbin_bssdefault,
  rawbin_targetlink,
  rawbin_lnksym,
  rawbin_setlnksym,
  rawbin_writeobject,
  rawbin_writeshared,
  srec37_write,
  0,
  2,
  0x7ff0,
  NULL,
  0,
  0,
  -1 /* endianess undefined, only write */
};
#endif



/*****************************************************************/
/*                        Read Binary                            */
/*****************************************************************/


static int rawbin_identify(char *name,uint8 *p,unsigned long plen)
/* identify a binary file */
{
  return (ID_UNKNOWN);  /* binaries are only allowed to be written! */
}


static void rawbin_readconv(struct GlobalVars *gv,struct LinkFile *lf)
{
  ierror("rawbin_readconv(): Can't read raw-binaries");
}


static uint8 rawbin_cmpsecflags(uint8 oldflags,uint8 newflags)
/* compare and verify target-specific section flags, */
/* return 0xff if sections are incompatible, otherwise return new flags */
{
  if ((oldflags&SF_UNINITIALIZED) != (newflags&SF_UNINITIALIZED))
    return (0xff);
  return (oldflags|newflags);
}


static struct Section *rawbin_bssdefault(struct ObjectUnit *ou)
/* Create a default BSS section for raw binaries */
{
  static char *bssname = ".bss";
  struct Section *s = create_section(ou,bssname,NULL,0);

  s->flags |= SF_UNINITIALIZED;
  s->type = ST_UDATA;
  s->protection = SP_READ | SP_WRITE;
  s->alignment = 2;
  return (s);
}


static int rawbin_targetlink(struct GlobalVars *gv,struct LinkedSection *ls,
                             struct Section *s)
/* returns 1, if target requires the combination of the two sections, */
/* returns -1, if target don't want to combine them, */
/* returns 0, if target doesn't care - standard linking rules are used. */
{
  ierror("rawbin_targetlink(): Impossible to link raw-binaries");
  return (0);
}


static struct Symbol *rawbin_lnksym(struct GlobalVars *gv,
                                    struct Section *sec,
                                    struct XReference *xref)
{
  return (NULL);
}


static void rawbin_setlnksym(struct GlobalVars *gv,struct Symbol *xdef,
                             struct XReference *xref)
{
}



/*****************************************************************/
/*                        Write Binary                           */
/*****************************************************************/


static void rawbin_writeexec(struct GlobalVars *gv,FILE *f,bool singlefile)
/* creates executable raw-binary files (with absolute addresses) */
{
  FILE *firstfile = f;
  bool firstsec = TRUE;
  unsigned long addr = ~0;
  struct LinkedSection *ls;

  while (ls = get_next_section(gv)) {
    if (ls->size == 0)
      continue;  /* ignore empty sections */

    if (firstsec) {
      if (gv->trace_file)
        fprintf(gv->trace_file,"Base address = 0x%08lx.\n",ls->copybase);
      firstsec = FALSE;
    }
    else {
      if (ls->copybase != addr) {
        if (ls->copybase-addr < MAXGAP || singlefile) {
          writegap(f,ls->copybase-addr);
        }
        else {  /* open a new file for this section */
          char *name;

          if (f != firstfile)
            fclose(f);
          name = alloc(strlen(gv->dest_name)+32);
          sprintf(name,"%s.%lx",gv->dest_name,ls->copybase);
          if (!(f = fopen(name,"w"))) {
            error(29,name);
            break;
          }
          free(name);
        }
      }
    }

    /* resolve all relocations and write section contents */
    calc_relocs(gv,ls);
    fwritex(f,ls->data,ls->filesize);
    if (ls->filesize < ls->size)
      writegap(f,ls->size - ls->filesize);

    addr = ls->copybase + ls->size;
  }

  if (f && f!=firstfile)
    fclose(f);
}


static void calc_relocs(struct GlobalVars *gv,struct LinkedSection *ls)
/* calculate all relocations and insert them using current endianess */
{
  signed char en = gv->big_endian;
  struct Reloc *rel;

  while (rel = (struct Reloc *)remhead(&ls->relocs)) {
    uint8 *p = ls->data + rel->offset;
    int32 x = (int32)rel->relocsect.lnk->base + rel->addend;

    switch (rel->type) {
      case R_NONE:
        break;

      case R_ADDR32:
        write32(en,p,(uint32)x);
        break;

      case R_ADDR26:
        write32(en,p,(read32(en,p)&0xfc000003) | (((uint32)x)&0x3fffffc));
        break;

      case R_ADDR16:
        write16(en,p,(uint16)x);
        break;

      case R_ADDR16_LO:
        write16(en,p,(uint16)(x & 0xffff));
        break;

      case R_ADDR16_HI:
        write16(en,p,(uint16)(x >> 16));
        break;

      case R_ADDR16_HA:
        write16(en,p,(uint16)(x >> 16) + ((x&0x8000)?1:0));
        break;

      /* @@@ currently no differentiation between the following */
      case R_ADDR14:
      case R_ADDR14_BRTAKEN:
      case R_ADDR14_BRNTAKEN:
        write16(en,p,(read16(en,p)&3) | (((uint16)x)&0xfffc));
        break;

      case R_ADDR8:
        *p = (uint8)x;
        break;

      default:
        ierror("calc_relocs(): Unsupported absolute relocation "
               "type %d",(int)rel->type);
        break;
    }
  }
}


static struct LinkedSection *get_next_section(struct GlobalVars *gv)
/* returns pointer to next section with lowest base address */
{
  struct LinkedSection *ls = (struct LinkedSection *)gv->lnksec.first;
  struct LinkedSection *nextls,*minls = NULL;

  while (nextls = (struct LinkedSection *)ls->n.next) {
    if (minls) {
      if (ls->copybase < minls->copybase)
        minls = ls;
    }
    else
      minls = ls;
    ls = nextls;
  }
  if (minls) {
    remnode(&minls->n);
  }
  return (minls);
}


static void writegap(FILE *f,long bytes)
{
  uint8 buf[GAPBUFSIZE];

  memset(buf,0,GAPBUFSIZE);
  do
    fwritex(f,buf,bytes>GAPBUFSIZE?GAPBUFSIZE:bytes);
  while ((bytes-=GAPBUFSIZE) > 0);
}


static void rawbin_writeshared(struct GlobalVars *gv,FILE *f)
{
  error(30);  /* Target file format doesn't support shared objects */
}


static void rawbin_writeobject(struct GlobalVars *gv,FILE *f)
{
  error(62);  /* Target file format doesn't support relocatable objects */
}


#ifdef RAWBIN1
static void rawbin_writesingle(struct GlobalVars *gv,FILE *f)
/* creates a single raw-binary file, fill gaps between */
/* sections with zero */
{
  rawbin_writeexec(gv,f,TRUE);
}
#endif


#ifdef RAWBIN2
static void rawbin_writemultiple(struct GlobalVars *gv,FILE *f)
/* creates raw-binary which might get splitted over several */
/* files, because of different section base addresses */
{
  rawbin_writeexec(gv,f,FALSE);
}
#endif


#if defined(SREC19) || defined(SREC28) || defined(SREC37)
static void SRecOut(FILE *f,int stype,uint8 *buf,int len)
{
  uint8 chksum = 0xff-len-1;

  fprintf(f,"S%1d%02X",stype,(unsigned)(len+1));
  for (; len>0; len--) {
    fprintf(f,"%02X",(unsigned)*buf);
    chksum -= *buf++;
  }
  fprintf(f,"%02X\n",(unsigned)chksum);
}


static void srec_write(struct GlobalVars *gv,FILE *f,int addrsize)
{
  bool firstsec = TRUE;
  struct LinkedSection *ls;
  unsigned long len,addr;
  uint8 *p,buf[MAXSREC+8];

  /* write header */
  buf[0] = buf[1] = 0;
  strncpy(&buf[2],gv->dest_name,MAXSREC+6);
  SRecOut(f,0,buf,(strlen(gv->dest_name)<(MAXSREC+6)) ?
          strlen(gv->dest_name)+2 : MAXSREC+8);

  while (ls = get_next_section(gv)) {
    if (ls->size == 0)
      continue;  /* ignore empty sections */

    if (firstsec) {
      if (gv->trace_file)
        fprintf(gv->trace_file,"Base address = 0x%08lx.\n",ls->copybase);
      firstsec = FALSE;
    }

    /* resolve all relocations and write section contents */
    calc_relocs(gv,ls);

    for (p=ls->data,addr=ls->copybase,len=ls->filesize; len>0; ) {
      int nbytes = (len>MAXSREC) ? MAXSREC : len;

      switch (addrsize) {
        case 2:
          write16be(buf,(uint16)addr);
          memcpy(buf+2,p,nbytes);
          SRecOut(f,1,buf,2+nbytes);
          break;
        case 3:
          buf[0] = (uint8)((addr>>16)&0xff);
          buf[1] = (uint8)((addr>>8)&0xff);
          buf[2] = (uint8)(addr&0xff);
          memcpy(buf+3,p,nbytes);
          SRecOut(f,2,buf,3+nbytes);
          break;
        case 4:
          write32be(buf,(uint32)addr);
          memcpy(buf+4,p,nbytes);
          SRecOut(f,3,buf,4+nbytes);
          break;
        default:
          ierror("srec_write(): Illegal SRec-type: %d",addrsize);
          nbytes = len;
          break;
      }
      p += nbytes;
      addr += nbytes;
      len -= nbytes;
    }
  }

  /* write trailer */
  memset(buf,0,4);
  SRecOut(f,11-addrsize,buf,addrsize);
}

#endif


#ifdef SREC19
static void srec19_write(struct GlobalVars *gv,FILE *f)
/* creates a Motorola S-Record file (S0,S1,S9), using 16-bit addresses */
{
  srec_write(gv,f,2);
}
#endif


#ifdef SREC28
static void srec28_write(struct GlobalVars *gv,FILE *f)
/* creates a Motorola S-Record file (S0,S2,S8), using 24-bit addresses */
{
  srec_write(gv,f,3);
}
#endif


#ifdef SREC37
static void srec37_write(struct GlobalVars *gv,FILE *f)
/* creates a Motorola S-Record file (S0,S1,S9), using 32-bit addresses */
{
  srec_write(gv,f,4);
}
#endif


#endif
